/* pepper.c */
/* pepper-monaka specific routines */

#define USE_8BIT_FTABLE
#ifndef INITIAL_CONFIG
#define INITIAL_CONFIG 7
#endif

struct portConfig_t {
  uchar ddrbIn;			/* DDRB 入力ピンのビット */
  uchar ddrbOut;		/* DDRB 出力ピンのビット */
} portConfig [] = {
  {0x39, 0x00}, /* Config 0 */
  {0x00, 0x39}, /* Config 1 */
  {0x20, 0x19}, /* Config 2 */
  {0x20, 0x19}, /* Config 3 */
  {0x39, 0x00}, /* Config 4 */
  {0x39, 0x00}, /* Config 5 */
  {0x39, 0x00}, /* Config 6 */
  {0x39, 0x00}, /* Config 7 */
};

struct portInfo_t {
  uchar tone_no;
  volatile uchar *port;
  uchar bit;
  uchar status;
};

/* ポートとトーンのマッピングテーブル(出力用) */
struct portInfo_t portInfoOut []= {
  {60, &PORTB, 3, 0},		/* C3 */
  {61, &PORTB, 4, 0},		/* C#3 */
  {62, &PORTB, 0, 0},		/* D3 */
  {63, &PORTB, 5, 0},		/* D#3 */
};

/* ポートとトーンのマッピングテーブル(Config5,6 入力用) */
/* その他のモードでは、portInfoIn[0]の組み込みボタンだけを読込む */
struct portInfo_t portInfoIn []= {
  {60, &PINB, 3, 0},		/* C3 */
  {61, &PINB, 4, 0},		/* C#3 */
  {62, &PINB, 0, 0},		/* D3 */
  {63, &PINB, 5, 0},		/* D#3 */
};

/* テーブル参照用 */
uchar polyOutMin = 60;
uchar polyOutMax = 62;

uchar scanKeyMin = 0;
uchar scanKeyMax = 4;

/*--------- MONOME MODE KEY SCAN用 ------------- */
struct scanPortInfo_t {
  volatile uchar *port;
  uchar bit;
  volatile uchar *ddr;
  volatile uchar *in;
};

#define SCAN_ROW_SIZE 2
#define SCAN_COL_SIZE 2

struct scanPortInfo_t scanCol []= {
  {&PORTB, 0, &DDRB, &PINB},
  {&PORTB, 5, &DDRB, &PINB},
};

struct scanPortInfo_t scanRow []= {
  {&PORTB, 3, &DDRB, &PINB},
  {&PORTB, 4, &DDRB, &PINB},
};

/*--------- MONOME MODE KEY SCAN用 ------------- */
/* インプリされたモードのチェック */
#define MODE_CHECK(c) (((c) < 4) || ((c) == 5) || ((c) == 7))

/*----------------------------------------------- */
/* I/O入出力LEDの処理 */

/* not supported on pepper */
#define SET_LED(val) ;

/* timing clock support */
#define CLOCK_OUT() { if (conf == 3) {flipPortBit(&PORTB, 0); } }

/* polyOutのあるモード */
#define POLY_OUT_CHECK() (conf <= 3)

/* trigOutのあるモード */
#define TRIG_OUT_CHECK() (conf == 3)

/* not supported on pepper */
#define monoOut(note, val) ;

/* for ToneOut */
#define SND_OUT_CHECK() (conf == 2)
#define SND_ON  {OCR0A=getTp(note); DDRB |=  0b00000001; } /* PB0=out */
#define SND_OFF {DDRB &= ~0b00000001; } /* PB0=in */

/* key input */
#define MAX_KEYS 4
static uchar keyCountButton[MAX_KEYS];
static uchar buttonStatus[MAX_KEYS];
/* matrix LED status */
static uchar matrixLedStatus[MAX_KEYS];

#define SCAN_KEY_CHECK() ((conf == 2) || (conf == 3) || (conf == 5))

#define KEY_MODE_READ(x) ;
#define KEY_EVENT_SEND(note, x) {			\
    keyEventAppend((note), x); \
  }

#define SCAN_MATRIX_CHECK() (conf == 7)

void portInit(uchar cf) {
  /* setup PORTB */
  DDRB &= ~portConfig[cf].ddrbIn;
  DDRB |= portConfig[cf].ddrbOut;
  PORTB |= portConfig[cf].ddrbIn;  /* pull-up all input ports */
  PORTB &= ~portConfig[cf].ddrbOut; /* PUTPUT 0 for all output pin */

  /* PWMの設定をする */

  if (cf == 2) {
    /* TC0, TC1を初期化 */
    TCCR0A = 0b01000010;	/* CTC mode, Toggle OC0A on Compare Match, OC0B Disconnected  */
    TCCR0B = 0b00000100;	/* 1/256 clock */
    OCR0A = 0;			/* for channel 0 */
  } else {
    TCCR0A = 0b00000010;	/* OC0A/OC0B DISCONNECTDC0B  */
    TCCR0B = 0b00000100;	/* 1/256 clock */
  }

  /* キースキャンテーブルの初期化 */
  initializeKeyStatus();

  /* テーブル参照用 */
  polyOutMin = 60;
  polyOutMax = polyOutMin + ((conf == 1) ? 4 : 2); /* 64: MODE=1, 62 : MODE=2,3 */

  scanKeyMin = (conf == 5) ? 0 : 3;
  scanKeyMax = 4;

  polyOutInit();
}

/* ------------------------------------------------------------------------- */
/* ------------------------ Oscillator Calibration ------------------------- */
/* ------------------------------------------------------------------------- */

/* Calibrate the RC oscillator to 8.25 MHz. The core clock of 16.5 MHz is
 * derived from the 66 MHz peripheral clock by dividing. Our timing reference
 * is the Start Of Frame signal (a single SE0 bit) available immediately after
 * a USB RESET. We first do a binary search for the OSCCAL value and then
 * optimize this value with a neighboorhod search.
 * This algorithm may also be used to calibrate the RC oscillator directly to
 * 12 MHz (no PLL involved, can therefore be used on almost ALL AVRs), but this
 * is wide outside the spec for the OSCCAL value and the required precision for
 * the 12 MHz clock! Use the RC oscillator calibrated to 12 MHz for
 * experimental purposes only!
 */
static void calibrateOscillator(void) {
  uchar step = 128;
  uchar trialValue = 0, optimumValue;
  int   x, optimumDev, targetValue = (unsigned)(1499 * (double)F_CPU / 10.5e6 + 0.5);

  /* do a binary search: */
  do{
    OSCCAL = trialValue + step;
    x = usbMeasureFrameLength();    /* proportional to current real frequency */
    if(x < targetValue)             /* frequency still too low */
      trialValue += step;
    step >>= 1;
  }while(step > 0);
  /* We have a precision of +/- 1 for optimum OSCCAL here */
  /* now do a neighborhood search for optimum value */
  optimumValue = trialValue;
  optimumDev = x; /* this is certainly far away from optimum */
  for(OSCCAL = trialValue - 1; OSCCAL <= trialValue + 1; OSCCAL++){
    x = usbMeasureFrameLength() - targetValue;
    if(x < 0)
      x = -x;
    if(x < optimumDev){
      optimumDev = x;
      optimumValue = OSCCAL;
    }
  }
  OSCCAL = optimumValue;
}
/*
Note: This calibration algorithm may try OSCCAL values of up to 192 even if
the optimum value is far below 192. It may therefore exceed the allowed clock
frequency of the CPU in low voltage designs!
You may replace this search algorithm with any other algorithm you like if
you have additional constraints such as a maximum CPU clock.
For version 5.x RC oscillators (those with a split range of 2x128 steps, e.g.
ATTiny25, ATTiny45, ATTiny85), it may be useful to search for the optimum in
both regions.
*/

static uchar __attribute__((section(".eeprom"))) calVal;

void    usbEventResetReady(void) {
  calibrateOscillator();
  eeprom_write_byte(&calVal, OSCCAL);   /* store the calibrated value in EEPROM */
}

void calibrate() {
  uchar i;
  uchar calibrationValue;

  calibrationValue = eeprom_read_byte(&calVal); /* calibration value from last time */
  if(calibrationValue != 0xff){
    OSCCAL = calibrationValue;
  }
  odDebugInit();
  usbDeviceDisconnect();
  for(i=0;i<20;i++){  /* 300 ms disconnect */
    _delay_ms(15);
  }
}

/* EOF */
